"""Collection of optional external algorithms."""
from . import (scSLAT,CEFCON,mofapy2,GNTD,spaceflow,STT,
               tosica,STAGATE_pyG,STAligner,spatrio,PROST,cytotrace2,
               GraphST,commot,cnmf,starfysh,scMulan,flowsig,PyWGCNA,
               CAST,scanorama,scdiffusion,BINARY,cellanova,VIA,gaston
               )
__all__ = [
    'scSLAT',
    'CEFCON',
    'mofapy2',
    'GNTD',
    'spaceflow',
    'STT',
    'tosica',
    'STAGATE_pyG',
    'STAligner',
    'spatrio',
    'PROST',
    'cytotrace2',
    'GraphST',
    'commot',
    'cnmf',
    'starfysh',
    'scMulan',
    'flowsig',
    'PyWGCNA',
    'CAST',
    'scanorama',
    'scdiffusion',
    'BINARY',
    'cellanova',
    'VIA',
    'gaston',
]
